use std::{collections::VecDeque, sync::Arc};

use btleplug::{
    api::{Central as _, Peripheral as _},
    platform::PeripheralId,
};
pub use state::{CentralState, ChannelState, PeripheralItem};
use tauri::{async_runtime::RwLock, AppHandle, Emitter, State};

use crate::{
    bluetooth, errors,
    pattern::Pattern,
    state::{self as global, AppState},
};

mod state;

#[tauri::command]
pub async fn activate_central_adapter(
    app_handle: AppHandle,
    app_state: State<'_, Arc<RwLock<AppState>>>,
) -> Result<(), errors::AppError> {
    let app = Arc::new(app_handle);
    let adapter = bluetooth::get_central_adapter().await?;
    {
        let state = app_state.read().await;
        global::set_central_adapter(adapter).await;
        state.clear_central_event_handler().await;
    }
    let handle = bluetooth::handle_bluetooth_events(Arc::clone(&app)).await?;
    {
        let state = app_state.write().await;
        state.set_central_event_handler(handle).await;
    }
    app.emit("central_state_updated", ()).unwrap();
    Ok(())
}

#[tauri::command]
pub async fn central_device_state() -> Result<CentralState, errors::AppError> {
    let central = global::get_central_adapter().await;
    if let Some(central) = central {
        let central_device_state = central
            .adapter_state()
            .await
            .map_err(|_| errors::AppError::BluetoothNotReady)?;
        Ok(CentralState {
            is_ready: central_device_state == btleplug::api::CentralState::PoweredOn,
            is_scanning: global::get_scanning().await,
            connected: global::get_connected_peripheral().await,
        })
    } else {
        Ok(CentralState::default())
    }
}

#[tauri::command]
pub async fn connected_peripheral_state() -> Result<Option<PeripheralItem>, errors::AppError> {
    let connected_peripheral = global::get_connected_peripheral()
        .await
        .clone()
        .ok_or(errors::AppError::NoConnectedPeripheral)?;
    let central = global::get_central_adapter()
        .await
        .ok_or(errors::AppError::BluetoothAdapterNotFound)?;
    if let Ok(peripheral) = central.peripheral(&connected_peripheral).await {
        let properties = peripheral
            .properties()
            .await
            .map_err(|_| errors::AppError::UnableToRetrievePeripheralProperties)?;
        peripheral
            .discover_services()
            .await
            .map_err(|_| errors::AppError::UnableToRetrievePeripheralServices)?;
        Ok(Some(
            PeripheralItem::new_form_peripheral(&peripheral, properties.as_ref()).await,
        ))
    } else {
        Err(errors::AppError::NoConnectedPeripheral)
    }
}

#[tauri::command]
pub async fn specific_peripheral_state(
    id: PeripheralId,
) -> Result<PeripheralItem, errors::AppError> {
    let central = global::get_central_adapter()
        .await
        .ok_or(errors::AppError::BluetoothAdapterNotFound)?;
    let peripheral = central
        .peripheral(&id)
        .await
        .map_err(|_| errors::AppError::UnableToRetrievePeripheralState)?;
    let properties = peripheral.properties().await.unwrap_or(None);
    let peripheral_item = if peripheral.connect().await.is_ok() {
        peripheral.discover_services().await.unwrap_or_default();
        let item = PeripheralItem::new_form_peripheral(&peripheral, properties.as_ref()).await;
        peripheral.disconnect().await.unwrap();
        item
    } else {
        let services = global::get_found_setvices(&id).await;
        PeripheralItem::new_with_cache(&peripheral, properties.as_ref(), services).await
    };
    Ok(peripheral_item)
}

#[tauri::command]
pub async fn channel_a_state(
    _app_state: State<'_, Arc<RwLock<AppState>>>,
) -> Result<ChannelState, errors::AppError> {
    Ok(ChannelState::default())
}

#[tauri::command]
pub async fn channel_b_state(
    _app_state: State<'_, Arc<RwLock<AppState>>>,
) -> Result<ChannelState, errors::AppError> {
    Ok(ChannelState::default())
}

#[tauri::command]
pub async fn start_scan_devices(app_handle: AppHandle) -> Result<(), errors::AppError> {
    bluetooth::start_scan(Arc::new(app_handle)).await?;
    Ok(())
}

#[tauri::command]
pub async fn stop_scan_devices(app_handle: AppHandle) -> Result<(), errors::AppError> {
    bluetooth::stop_scan(Arc::new(app_handle)).await?;
    Ok(())
}

#[tauri::command]
pub async fn list_patterns(
    app_state: State<'_, Arc<RwLock<AppState>>>,
    keyword: Option<String>,
) -> Result<VecDeque<Pattern>, errors::AppError> {
    let state = app_state.read().await;
    let patterns = state
        .db
        .get_patterns(keyword)
        .await
        .map_err(|e| errors::AppError::StorageFailure(e.to_string()))?;
    Ok(patterns)
}
